
/*
 * list.c
 *
 * (C) Copyright 1995 Archie L. Cobbs
 */

#include <stdio.h>
#include <assert.h>
#include "util.h"
#include "error.h"
#include "names.h"
#include "obj.h"
#include "asm.h"

/*
 * INTERNAL FUNTIONS
 */

	static	void		ShowLineNum(FILE *fp);
	static	void		ShowLoc(FILE *fp, int segment, unsigned address);

/*
 * ListLine()
 *
 * Assumes no trailing newline in gLineBuf
 */

void
ListLine(FILE *fp)
{
	int				nbytes, extra, k;
	static int		pass = 0;
	static int		segment;
	static unsigned	address;

/* Make sure we know where we are */

	if (pass != gPass)
	{
		segment = INITIAL_SEGMENT;
		address = 0;
		pass = gPass;
	}

/* Show line number */

	ShowLineNum(fp);

/* Show location for label or if bytes were emitted */

	if (gLocLabel || address != gAddress[segment])
		ShowLoc(fp, segment, address);
	else
		fprintf(fp, "\t");

/* Compute number of total and expanded bytes to show */

	if (segment == BssSegment)
		nbytes = 0;
	else
		nbytes = gAddress[segment] - address;

	if (nbytes > 4)
	{
		extra = nbytes - 4;
		nbytes = 4;
	}
	else
		extra = 0;

/* Always show up to four emitted bytes */

	for (k = 0; k < nbytes; k++)
	{
		fprintf(fp, "%02X", gSegBuffer[segment][address++]);
		if (k < nbytes - 1)
			fprintf(fp, " ");
	}
	if (nbytes <= 1)
		fprintf(fp, "\t");
	fprintf(fp, "\t");

/* Now show the input line */

	fprintf(fp, "%s\n", gLineBuf);

/* Show remaining bytes if expand is on */

	if (gExpand && segment != BssSegment)
		while (extra)
		{
			ShowLineNum(fp);
			ShowLoc(fp, segment, address);
			nbytes = extra;
			nbytes = (nbytes <= 8) ? nbytes : 8;
			extra -= nbytes;
			for (k = 0; k < nbytes; k++)
			{
				fprintf(fp, "%02X", gSegBuffer[segment][address++]);
				if (k < nbytes - 1)
					fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

	fflush(fp);

/* Remember where we are now for next time */

	segment = gSegment;
	address = gAddress[gSegment];
}

/*
 * ShowLineNum()
 */

static	void
ShowLineNum(FILE *fp)
{
	fprintf(fp, "%4d", gLineNumber);
	if (gLineNumber < 10000)
		fprintf(fp, " ");
}

/*
 * ShowLoc()
 */

static	void
ShowLoc(FILE *fp, int segment, unsigned address)
{
	fprintf(fp, "%c %04X ",
		segment == CodeSegment ? 'C' :
		segment == DataSegment ? 'D' :
		'B', address);
}

