
/*
 * ops.c
 *
 * (C) Copyright 1995 Archie L. Cobbs
 */

#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "error.h"
#include "names.h"
#include "asm.h"

/*
 * DEFINITIONS
 */

	#define	NUM_MODES		12

	OpaqueDecl(OpInfo);
	OpaqueDefn(OpInfo)
	{
		unsigned	id;
		Byte		ops[NUM_MODES];
	};

/*
 * INTERNAL FUNCTIONS
 */

	static	OpInfo		Locate(unsigned iname);

/*
 * Instructions and opcodes
 *
 * NOTE: We use $00 to mean "no opcode" and treat BRK as special case
 */

	#define	OP(ins,imp,rel,imm,zpg,zpx,zpy,abs,abx,aby,inx,iny,ind)	\
	{																\
		(unsigned) #ins,											\
		{															\
			0x##imp, 0x##rel, 0x##imm, 0x##zpg,						\
			0x##zpx, 0x##zpy, 0x##abs, 0x##abx,						\
			0x##aby, 0x##inx, 0x##iny, 0x##ind						\
		}															\
	},

	static	struct OpInfo	gOpInfo[] =
	{

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  BRK,	00,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

		OP(  CLC,	18,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  CLD,	D8,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  CLI,	58,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  CLV,	B8,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

		OP(  SEC,	38,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  SED,	F8,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  SEI,	78,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

		OP(  DEX,	ca,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  DEY,	88,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

		OP(  INX,	e8,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  INY,	c8,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  BCC,	 0, 90,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BLT,	 0, 90,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BCS,	 0, B0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BGE,	 0, B0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BEQ,	 0, F0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BMI,	 0, 30,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BNE,	 0, D0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BPL,	 0, 10,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  BVC,	 0, 50,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  NOP,	EA,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  RTI,	40,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  RTS,	60,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

		OP(  PHA,	48,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  PLA,	68,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  PHP,	08,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  PLP,	28,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

		OP(  TAX,	aa,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  TAY,	a8,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  TSX,	ba,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  TXA,	8a,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  TXS,	9a,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)
		OP(  TYA,	98,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  JMP,	 0,  0,  0,  0,  0,  0, 4c,  0,  0,  0,  0,  6c	)
		OP(  JSR,	 0,  0,  0,  0,  0,  0, 20,  0,  0,  0,  0,   0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  LDA,	 0,  0, a9, a5, b5,  0, ad, bd, b9, a1, b1,  0	)
		OP(  LDX,	 0,  0, a2, a6,  0, b6, ae,  0, be,  0,  0,  0	)
		OP(  LDY,	 0,  0, a0, a4, b4,  0, ac, bc,  0,  0,  0,  0	)
		OP(  STA,	 0,  0,  0, 85, 95,  0, 8d, 9d, 99, 81, 91,  0	)
		OP(  STX,	 0,  0,  0, 86,  0, 96, 8e,  0,  0,  0,  0,  0	)
		OP(  STY,	 0,  0,  0, 84, 94,  0, 8c,  0,  0,  0,  0,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  CMP,	 0,  0, c9, c5, d5,  0, cd, dd, d9, c1, d1,  0	)
		OP(  CPX,	 0,  0, e0, e4,  0,  0, ec,  0,  0,  0,  0,  0	)
		OP(  CPY,	 0,  0, c0, c4,  0,  0, cc,  0,  0,  0,  0,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  ADC,	 0,  0, 69, 65, 75,  0, 6d, 7d, 79, 61, 71,  0	)
		OP(  SBC,	 0,  0, e9, e5, f5,  0, ed, fd, f9, e1, f1,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  AND,	 0,  0, 29, 25, 35,  0, 2d, 3d, 39, 21, 31,  0	)
		OP(  ORA,	 0,  0, 09, 05, 15,  0, 0d, 1d, 19, 01, 11,  0	)
		OP(  EOR,	 0,  0, 49, 45, 55,  0, 4d, 5d, 59, 41, 51,  0	)
		OP(  XOR,	 0,  0, 49, 45, 55,  0, 4d, 5d, 59, 41, 51,  0	)
		OP(  BIT,	 0,  0,  0, 24,  0,  0, 2c,  0,  0,  0,  0,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(  ASL,	0a,  0,  0, 06, 16,  0, 0e, 1e,  0,  0,  0,  0	)
		OP(  ROL,	2a,  0,  0, 26, 36,  0, 2e, 3e,  0,  0,  0,  0	)
		OP(  LSR,	4a,  0,  0, 46, 56,  0, 4e, 5e,  0,  0,  0,  0	)
		OP(  ROR,	6a,  0,  0, 66, 76,  0, 6e, 7e,  0,  0,  0,  0	)

/******     INST,   Imp Rel Imm ZPg ZPX ZPY Abs AbX AbY InX InY Ind *******/

		OP(	 DEC,	 0,  0,  0, c6, d6,  0, ce, de,  0,  0,  0,  0	)
		OP(	 INC,	 0,  0,  0, e6, f6,  0, ee, fe,  0,  0,  0,  0	)

	};

	static	int		kNumInstr = sizeof(gOpInfo) / sizeof(*gOpInfo);

/*
 * OpcodeInit()
 *
 * Replace all opcode strings with string ID's
 * Note that ID's will be in increasing order,
 * so we can do a binary search.
 */

void
OpcodeInit(void)
{
	int		k;

	for (k = 0; k < kNumInstr; k++)
		gOpInfo[k].id = NameID(gNames, (char *) gOpInfo[k].id);
}

/*
 * GetOp()
 *
 * Get operand for instruction, address mode pair.
 * Generate error message if appropriate.
 */

Byte
GetOp(unsigned ins, int mode)
{
	int		opcode;

/* Check for valid instruction + address mode */

	if (Locate(ins) == NULL)
	{
		uerror("unknown instruction \"%s\"", Name(gNames, ins));
		return(0);
	}
	if ((opcode = CheckOp(ins, mode)) < 0)
	{
		uerror("invalid address mode");
		return(0);
	}

/* Done */

	return(opcode);
}

/*
 * CheckOp()
 *
 * Get operand for instruction, address mode pair.
 * Returns negative if invalid.
 */

int
CheckOp(unsigned ins, int mode)
{
	OpInfo	info;

/* Check for valid instruction */

	if ((info = Locate(ins)) == NULL)
		return(-1);

/* Return opcode or -1 if invalid mode */

	return((ins == NameID(gNames, "BRK")) ?
		((mode == mImplied) ? 0x00 : -1) :
		(info->ops[mode - 1] ? info->ops[mode - 1] : -1));
}

/*
 * Locate()
 *
 * Find the OpInfo structure for given instruction (using binary search)
 */

static	int
LocateCmp(const void *a, const void *b)
{
	return(((OpInfo) a)->id - ((OpInfo) b)->id);
}

static	OpInfo
Locate(unsigned iname)
{
	struct OpInfo	key;

	key.id = iname;
	return((OpInfo) bsearch(&key, gOpInfo,
		kNumInstr, sizeof(*gOpInfo), LocateCmp));
}

