
/*
 * parse.y
 *
 * Bison input file for Apple2 assembler
 * (C) Copyright 1995 Archie L. Cobbs
 *
 */

/***** C DECLARATIONS *****/

%{

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "util.h"
#include "error.h"
#include "names.h"
#include "obj.h"
#include "asm.h"

/* For error reporting */

	#define	yyerror		uerror

/* Generic constant expression evaluation */

	#define	UnaryOp(a,b,OPR)													\
	{																			\
		if (a.type == iUnknown)													\
			b.type = iUnknown;													\
		else if (a.type != iSegRel || a.u.segrel.segment != AbsSegment)			\
			uerror("illegal expression");										\
		else																	\
		{																		\
			b.type = iSegRel;													\
			b.u.segrel.segment = AbsSegment;									\
			b.u.segrel.offset = OPR b.u.segrel.offset;							\
		}																		\
	}

	#define	BinaryOp(a,b,c,OPR)													\
	{																			\
		if (a.type == iUnknown || b.type == iUnknown)							\
			c.type = iUnknown;													\
		else if (a.type != iSegRel || a.u.segrel.segment != AbsSegment			\
			  || b.type != iSegRel || b.u.segrel.segment != AbsSegment)			\
			uerror("illegal expression");										\
		else																	\
		{																		\
			c.type = iSegRel;													\
			c.u.segrel.segment = AbsSegment;									\
			c.u.segrel.offset = a.u.segrel.offset OPR b.u.segrel.offset;		\
		}																		\
	}

%}

/***** BISON DECLARATIONS *****/

/* Tokens, precedence */

	%left	','
	%right	'?' ':'
	%left	LOR
	%left	LAND
	%left	'|'
	%left	'^'
	%left	'&'
	%left	EQU NEQ
	%left	'<' '>' LE GE
	%left	LSH RSH
	%left	'+' '-'
	%left	'*' '/' '%'
	%right	'!' '~'

	%token	LSB MSB

	%token	ALIGN ASCII ASCIZ BEGNN BSS BYTE CODE
	%token	DATA DBYT DCI ELSE END ENDIF ENDM IF
	%token	EXPAND FLOAT GLOBAL HIBIT MACRO LONG
	%token	SPACE WORD ZERO IMPORT ZIMPORT

/* Semantic types */

	%union
	{
		unsigned		id;				/* Identifier */
		int				intval;			/* Integer literal */
		double			floatval;		/* Floating point literal */
		char			charval;		/* Character literal */
		struct TValue	textval;		/* String literal */
		struct IValue	ivalue;			/* Ivalue */
		struct List		ivlist;			/* List of ivalues */
		struct List		idlist;			/* List of identifiers */
		struct List		flist;			/* List of floats */
	}

/* Default semantic types */

	%token	<intval>	INTLIT
	%token	<floatval>	FLOATLIT
	%token	<charval>	CHARLIT
	%token	<textval>	TEXTLIT
	%token	<id>		LABEL IDENT

	%type	<intval>	const_int
	%type	<ivalue>	i_value
	%type	<textval>	t_value
	%type	<floatval>	f_value
	%type	<ivlist>	iv_list
	%type	<idlist>	id_list
	%type	<flist>		f_list

%%

/***** GRAMMER DEFINITION *****/

program:
	complete_line
	| program complete_line 
	;

complete_line:
	line '\n'
	{
		if (gListFile && gPass == 2)
			ListLine(gListFile);
		gLineNumber++;
		gLocLabel = FALSE;
	}
	;

line:
	unlabeled_line
	| local_label unlabeled_line
	| LABEL '=' i_value
	{
		Symbol	sym = GetSym($1);

	/* Check stuff */

		switch (gPass)
		{
			case 1:
				if (sym)
					uerror("symbol \"%s\" defined previously: "
						"file \"%s\", line %d", Name(gNames, $1),
						Name(gNames, sym->file), sym->line);
				else
				{
					sym = DefineSym($1);
					sym->file = NameID(gNames, gFileName);
					sym->line = gLineNumber;
				}
				break;
			case 2:
				if (!sym)
					cerror("lost symbol \"%s\"", Name(gNames, sym->id));
				break;
			default:
				assert(0);
		}

	/* Set value */

		switch ($3.type)
		{
			case iSegRel:
				sym->type = sSegRel;
				sym->u.segrel = $3.u.segrel;
				sym->valid = TRUE;
				break;
			case iSymOff:
				sym->type = sSymOff;
				sym->u.symoff = $3.u.symoff;
				sym->valid = TRUE;
				break;
			case iUnknown:
				if (gPass == 1)
				{
					sym->type = sSegRel;
					sym->u.segrel.segment = AbsSegment;
					sym->u.segrel.offset = 0;
				}
				else
					uerror("can't resolve this expression");
				break;
			default:
				assert(0);
		}

	/* Debugging */

		if (gDebug)
		{
			fprintf(stderr, "--> DEFINED: ");
			ShowSym(stderr, sym);
		}
	}
	;

local_label:
	LABEL
	{
		Symbol	sym = GetSym($1);

		gLocLabel = TRUE;
		switch (gPass)
		{
			case 1:
				if (sym)
					uerror("symbol \"%s\" defined previously: "
						"file \"%s\", line %d", Name(gNames, $1),
						Name(gNames, sym->file), sym->line);
				else
				{
					sym = DefineSym($1);
					sym->file = NameID(gNames, gFileName);
					sym->line = gLineNumber;
					sym->type = sSegRel;
					sym->u.segrel.segment = gSegment;
					sym->u.segrel.offset = gAddress[gSegment];
					sym->valid = TRUE;
					if (gDebug)
					{
						fprintf(stderr, "--> DEFINED: ");
						ShowSym(stderr, sym);
					}
				}
				break;

			case 2:
				if (!sym)
					cerror("symbol not found");
				if (!sym->valid || sym->type != sSegRel)
					cerror("symbol inconsistency");
				if (sym->u.segrel.segment != gSegment
					|| sym->u.segrel.offset != gAddress[gSegment])
					cerror("sync error");
				break;

			default:
				assert(0);
		}
	}
	;

unlabeled_line:
		/* empty */

/* Normal instruction operands */

	| IDENT									/* Implied */
	{
		EmitIns($1, NULL, mImplied, 0);
	}
	| IDENT LSB i_value						/* Immediate, low byte */
	{
		EmitImm($1, &$3, FALSE);
	}
	| IDENT MSB i_value						/* Immediate, high byte */
	{
		EmitImm($1, &$3, TRUE);
	}
	| IDENT i_value							/* Absolute, relative, or ZP */
	{
		EmitIns($1, &$2, mAbsolute, mZeroPage);
	}
	| IDENT '(' i_value ')'					/* Indirect */
	{
		EmitIns($1, &$3, mIndirect, 0);
	}
	| IDENT i_value ',' IDENT				/* Indexed */
	{
		int	mode, zmode;

		if ($4 == NameID(gNames, "x") || $4 == NameID(gNames, "X"))
			mode = mAbsoluteX, zmode = mZeroPageX;
		else if ($4 == NameID(gNames, "y") || $4 == NameID(gNames, "Y"))
			mode = mAbsoluteY, zmode = mZeroPageY;
		else
		{
			uerror("bad addressing mode");
			mode = mAbsoluteX, zmode = mZeroPageX;
		}
		EmitIns($1, &$2, mode, zmode);
	}
	| IDENT '(' i_value ')' ',' IDENT
	{
		if ($6 != NameID(gNames, "y") && $6 != NameID(gNames, "Y"))
			uerror("bad addressing mode");
		else
			EmitIns($1, &$3, 0, mIndirectY);
	}
	| IDENT '(' i_value ',' IDENT ')'
	{
		if ($5 != NameID(gNames, "x") && $5 != NameID(gNames, "X"))
			uerror("bad addressing mode");
		else
			EmitIns($1, &$3, 0, mIndirectX);
	}

/* .align */

	| ALIGN const_int			{ gAddress[gSegment] =
									RoundUp(gAddress[gSegment], $2); }

/* .code, .data, .bss */

	| CODE						{ gSegment = CodeSegment; }
	| DATA						{ gSegment = DataSegment; }
	| BSS						{ gSegment = BssSegment; }

/* .hibit, .expand */

	| HIBIT const_int			{ gHiBit = $2 ? 0x80 : 0x00; }
	| EXPAND const_int			{ gExpand = $2 ? TRUE : FALSE; }

/* .ascii .asciz .dci */

	| ASCII t_value				{ PsAscii(&$2, FALSE, FALSE); free($2.chars); }
	| ASCIZ t_value				{ PsAscii(&$2, TRUE, FALSE); free($2.chars); }
	| DCI t_value				{ PsAscii(&$2, FALSE, TRUE); free($2.chars); }

/* .byte, .word, .dbyt, .long */

	| BYTE iv_list				{ PsData(&$2, 1, FALSE); }
	| WORD iv_list				{ PsData(&$2, 2, FALSE); }
	| DBYT iv_list				{ PsData(&$2, 2, TRUE); }
	| LONG iv_list				{ PsData(&$2, 4, FALSE); }

/* .space */

	| SPACE const_int
	{
		if ($2 < 0 || $2 > 0x8000)
			uerror("value out of range");
		gAddress[gSegment] += $2;
	}

/* .zero */

	| ZERO const_int
	{
		int		k;

		if ($2 < 0 || $2 > 0x8000)
			uerror("value out of range");
		for (k = 0; k < $2; k++)
			Emit(0);
	}

/* .float */

	| FLOAT f_list
	{ 
		int			k;
		List		list, next;
		Byte		value[FLOAT_SIZE_6502];

	/* Emit floats */

		for (list = &$2; list; list = list->next)
		{
			EncodeFloat(list->u.fvalue, value);
			for (k = 0; k < sizeof(value); k++)
				Emit(value[k]);
		}

	/* Free tail of list */

		for (list = $2.next; list; list = next)
		{
			next = list->next;
			free(list);
		}
	}

/* .import and .zimport */

	| IMPORT id_list
	{
		Symbol	sym;
		List	list, next;

	/* Import symbols during pass one, verify during pass two */

		for (list = &$2; list; list = list->next)
			switch (gPass)
			{
				case 1:
					Import(list->u.id);
					break;
				case 2:
					if ((sym = GetSym(list->u.id)))
						uerror("symbol \"%s\" defined locally",
							Name(gNames, list->u.id));
					break;
				default:
					assert(0);
			}

	/* Free tail of list */

		for (list = $2.next; list; list = next)
		{
			next = list->next;
			free(list);
		}
	}
	| ZIMPORT id_list
	{
		Symbol	sym;
		List	list, next;

	/* Import symbols during pass one, verify during pass two */

		for (list = &$2; list; list = list->next)
			switch (gPass)
			{
				case 1:
					ZpImport(list->u.id);
					break;
				case 2:
					if ((sym = GetSym(list->u.id)))
						uerror("symbol \"%s\" defined locally",
							Name(gNames, list->u.id));
					break;
				default:
					assert(0);
			}

	/* Free tail of list */

		for (list = $2.next; list; list = next)
		{
			next = list->next;
			free(list);
		}
	}

/* .global == .export */

	| GLOBAL id_list
	{
		Symbol	sym;
		List	list, next;

	/* Export symbols during pass two */

		if (gPass == 2)
			for (list = &$2; list; list = list->next)
				if ((sym = GetSym(list->u.id)))
					Export(sym);
				else
					uerror("undefined symbol \"%s\"",
						Name(gNames, list->u.id));

	/* Free tail of list */

		for (list = $2.next; list; list = next)
		{
			next = list->next;
			free(list);
		}
	}

/* unimplementeds */

	| BEGNN						{ uerror("pseudo-op unimplemented"); }
	| END						{ uerror("pseudo-op unimplemented"); }
	| MACRO						{ uerror("pseudo-op unimplemented"); }
	| ENDM						{ uerror("pseudo-op unimplemented"); }
	| IF						{ uerror("pseudo-op unimplemented"); }
	| ELSE						{ uerror("pseudo-op unimplemented"); }
	| ENDIF						{ uerror("pseudo-op unimplemented"); }
	;

/* Lists of things -- note these are backwards, reverse ### */

id_list:
	IDENT
	{
		$$.u.id = $1;
		$$.next = NULL;
	}
	| IDENT ',' id_list
	{
		$$.u.id = $1;
		$$.next = NewOpaque(List);
		$$.next->u.id = $3.u.id;
	 	$$.next->next = $3.next;
	}
	;

iv_list:
	i_value
	{
		$$.u.ivalue = $1;
		$$.next = NULL;
	}
	| i_value ',' iv_list
	{
		$$.u.ivalue = $1;
		$$.next = NewOpaque(List);
		$$.next->u.ivalue = $3.u.ivalue;
		$$.next->next = $3.next;
	}
	;

f_list:
	f_value
	{
		$$.u.fvalue = $1;
		$$.next = NULL;
	}
	| f_value ',' f_list
	{
		$$.u.fvalue = $1;
		$$.next = NewOpaque(List);
		$$.next->u.fvalue = $3.u.fvalue;
		$$.next->next = $3.next;
	}
	;

i_value:
	const_int
	{
		$$.type = iSegRel;
		$$.u.segrel.segment = AbsSegment;
		$$.u.segrel.offset = $1;
	}
	| '[' i_value ']'
	{
		$$ = $2;
	}
	| const_int '?' i_value ':' i_value
	{
		$$ = $1 ? $3 : $5;
	}
	| '*'
	{
		$$.type = iSegRel;
		$$.u.segrel.segment = gSegment;
		$$.u.segrel.offset = gAddress[gSegment];
	}
	| IDENT
	{
		Symbol	sym = GetSym($1);

		switch (gPass)
		{
			case 1:
				if (sym)
					switch (sym->type)
					{
						case sSegRel:
							$$.type = iSegRel;
							$$.u.segrel = sym->u.segrel;
							break;
						case sSymOff:
							$$.type = iSymOff;
							$$.u.symoff = sym->u.symoff;
							break;
						default:
							assert(0);
					}
				else
					if (IsZpImport($1))
					{
						$$.type = iSymOff;
						$$.u.symoff.id = $1;
						$$.u.symoff.offset = 0;
					}
					else
						$$.type = iUnknown;
				break;

			case 2:
				if (sym)
					if (!sym->valid)
					{
						uerror("illegal forward reference to \"%s\"",
							Name(gNames, sym->id));
						$$.type = iSegRel;
						$$.u.segrel.segment = AbsSegment;
						$$.u.segrel.offset = 0;
					}
					else
						switch (sym->type)
						{
							case sSegRel:
								$$.type = iSegRel;
								$$.u.segrel = sym->u.segrel;
								break;
							case sSymOff:
								$$.type = iSymOff;
								$$.u.symoff = sym->u.symoff;
								break;
							default:
								assert(0);
						}
				else
				{
					$$.type = iSymOff;
					$$.u.symoff.id = $1;
					$$.u.symoff.offset = 0;
				}
				break;

			default:
				assert(0);
		}
	}

	| i_value '+' i_value
	{
		if ($1.type == iUnknown || $3.type == iUnknown)
			$$.type = iUnknown;
		else
			switch ($1.type)
			{
				case iSegRel:
					switch ($3.type)
					{
						case iSegRel:
							$$.type = iSegRel;
							if ($1.u.segrel.segment == AbsSegment)
								$$ = $3;
							else if ($3.u.segrel.segment == AbsSegment)
								$$ = $1;
							else
								uerror("illegal expression");
							$$.u.segrel.offset =
								$1.u.segrel.offset + $3.u.segrel.offset;
							break;
						case iSymOff:
							if ($1.u.segrel.segment != AbsSegment)
								uerror("illegal unresolved expression");
							$$ = $3;
							$$.u.symoff.offset += $1.u.segrel.offset;
							break;
						default:
							assert(0);
					}
					break;
				case iSymOff:
					switch ($3.type)
					{
						case iSegRel:
							if ($3.u.segrel.segment != AbsSegment)
								uerror("illegal unresolved expression");
							$$ = $1;
							$$.u.symoff.offset += $3.u.segrel.offset;
							break;
						case iSymOff:
							uerror("illegal unresolved expression");
							break;
					}
					break;
				default:
					assert(0);
			}
	}

	| i_value '-' i_value
	{
		if ($1.type == iUnknown || $3.type == iUnknown)
			$$.type = iUnknown;
		else
			switch ($3.type)
			{
				case iSegRel:
					switch ($1.type)
					{
						case iSegRel:
							if ($1.u.segrel.segment == $3.u.segrel.segment)
							{
								$$.type = iSegRel;
								$$.u.segrel.segment = AbsSegment;
								$$.u.segrel.offset =
									$1.u.segrel.offset - $3.u.segrel.offset;
							}
							else
							{
								if ($3.u.segrel.segment != AbsSegment)
									uerror("illegal expression");
								$$ = $1;
								$$.u.segrel.offset -= $3.u.segrel.offset;
							}
							break;
						case iSymOff:
							if ($3.u.segrel.segment != AbsSegment)
								uerror("illegal expression");
							$$ = $1;
							$$.u.symoff.offset -= $3.u.segrel.offset;
							break;
						default:
							assert(0);
					}
					break;
				case iSymOff:
					if ($1.type != iSymOff || $1.u.symoff.id != $3.u.symoff.id)
						uerror("illegal expression");
					$$.type = iSegRel;
					$$.u.segrel.segment = AbsSegment;
					$$.u.segrel.offset =
						$1.u.symoff.offset - $3.u.symoff.offset;
					break;
				default:
					assert(0);
			}
	}
	| '+' i_value					{ UnaryOp($2,$$,+); }
	| '-' i_value					{ UnaryOp($2,$$,-); }
	| '!' i_value					{ UnaryOp($2,$$,!); }
	| '~' i_value					{ UnaryOp($2,$$,~); }
	| i_value '*' i_value			{ BinaryOp($1,$3,$$,*); }
	| i_value '/' i_value			{ BinaryOp($1,$3,$$,/); }
	| i_value '%' i_value			{ BinaryOp($1,$3,$$,%); }
	| i_value LSH i_value			{ BinaryOp($1,$3,$$,<<); }
	| i_value RSH i_value			{ BinaryOp($1,$3,$$,>>); }
	| i_value LAND i_value			{ BinaryOp($1,$3,$$,&&); }
	| i_value LOR i_value			{ BinaryOp($1,$3,$$,||); }
	| i_value '&' i_value			{ BinaryOp($1,$3,$$,&); }
	| i_value '^' i_value			{ BinaryOp($1,$3,$$,^); }
	| i_value '|' i_value			{ BinaryOp($1,$3,$$,|); }
	| i_value EQU i_value			{ BinaryOp($1,$3,$$,==); }
	| i_value NEQ i_value			{ BinaryOp($1,$3,$$,!=); }
	| i_value '<' i_value			{ BinaryOp($1,$3,$$,<); }
	| i_value LE i_value			{ BinaryOp($1,$3,$$,<=); }
	| i_value '>' i_value			{ BinaryOp($1,$3,$$,>); }
	| i_value GE i_value			{ BinaryOp($1,$3,$$,>=); }
	;

const_int:
	INTLIT							{ $$ = $1; }
	| CHARLIT						{ $$ = $1; }
	;

f_value:
	i_value
	{
		if ($1.type != iSegRel || $1.u.segrel.segment != AbsSegment)
			uerror("illegal floating point expression");
		$$ = (double) $1.u.segrel.offset;
	}
	| FLOATLIT						{ $$ = $1; }
	| '+' FLOATLIT					{ $$ = $2; }
	| '-' FLOATLIT					{ $$ = -$2; }
	;

t_value:
	TEXTLIT
	| TEXTLIT TEXTLIT
	{
		$$.chars = GetMem($$.len = $1.len + $2.len);
		memcpy($$.chars, $1.chars, $1.len);
		memcpy($$.chars + $1.len, $2.chars, $2.len);
		free($1.chars);
		free($2.chars);
	}
	;

